from typing import (
    TYPE_CHECKING,
    Tuple,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    PP_AUTONEG,
    PP_AUTONEGSTATUS,
    PP_LINKTRAIN,
    PP_LINKTRAINSTATUS,
)


class AutoNeg:
    """L23 high-speed port PCS/PMA auto-negotiation"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.settings = PP_AUTONEG(conn, module_id, port_id)
        """Auto-negotiation settings of the PHY.
        
        :type: PP_AUTONEG
        """

        self.status = PP_AUTONEGSTATUS(conn, module_id, port_id)
        """Status of auto-negotiation settings of the PHY.
        
        :type: PP_AUTONEGSTATUS
        """


class LinkTrain:
    """L23 high-speed port PCS/PMA link training"""

    def __init__(self, conn: "itf.IConnection", port) -> None:
        self.settings = PP_LINKTRAIN(conn, *port.kind)
        """Link training settings.
        
        :type: PP_LINKTRAIN
        """

        self.per_lane_status: Tuple[PP_LINKTRAINSTATUS, ...] = tuple(
            PP_LINKTRAINSTATUS(conn, *port.kind, _serdes_xindex=idx)
            for idx in range(port.info.capabilities.serdes_count)
        ) 
        """Link training status.
        
        :type: PP_LINKTRAINSTATUS
        """


class PcsPma:
    """L23 high-speed port PCS/PMA settings"""

    def __init__(self, conn: "itf.IConnection", port) -> None:
        self.auto_neg = AutoNeg(conn, *port.kind)
        """PCS/PMA auto-negotiation settings.
        
        :type: AutoNeg
        """

        self.link_training = LinkTrain(conn, port)
        """PCS/PMA link training settings.
        
        :type: LinkTrain
        """


# # Temporary ports are not supporting LinkTrain, in future release of xenaserver it will be the same as regular PcsPma
# class PcsPmaL1:
#     """L23 high-speed port PCS/PMA settings"""

#     def __init__(self, conn: "itf.IConnection", port) -> None:
#         self.auto_neg = AutoNeg(conn, *port.kind)
#         """PCS/PMA auto-negotiation settings.
        
#         :type: AutoNeg
#         """



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum, IntEnum
import os
import sys
from typing import List
from ipaddress import (
    IPv4Address,
    IPv6Address,
)
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from xoa_driver.internals.core.transporter.protocol.payload.exceptions import FirmwareVersionError  # noqa: E402
from xoa_driver.internals.core.transporter.protocol.payload import (  # noqa: E402
    ResponseBodyStruct,
    field,
    XmpByte,
    XmpHex,
    XmpInt,
    XmpIPv4Address,
    XmpIPv6Address,
    XmpLong,
    XmpMacAddress,
    XmpSequence,
    XmpShort,
    XmpStr,
    Hex
)


# region TypeTesting

def test_byte() -> None:
    data = b"\xff\xff"
    assert len(data) == 2, "for current test case the data must be of 2 bytes long"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: int = field(XmpByte())
        custom_field2: int = field(XmpByte(signed=True))
        custom_field3: int = field(XmpByte())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, int)
    assert obj.custom_field == 255
    assert obj.custom_field != 65535
    assert obj.custom_field2 == -1
    try:
        obj.custom_field3
    except FirmwareVersionError:
        pass
    else:
        assert False, "Suppose to break on not suficient amount of bytes"


def test_short() -> None:
    data = b"\x02\x05"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: int = field(XmpShort())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, int)
    assert obj.custom_field != 2
    assert obj.custom_field != 255
    assert obj.custom_field == 517


def test_int() -> None:
    data = b"\x00\x00\x00\xff"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: int = field(XmpInt(climb=(10, 20)))

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, int)
    assert obj.custom_field != 0
    assert obj.custom_field == 255


def test_long() -> None:
    data = b"\x00\x00\x00\xff\x00\x00\x00\xff"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: int = field(XmpLong())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, int)
    assert obj.custom_field == 1095216660735


def test_hex() -> None:
    data = b"\x00\x00\x00\xff\x00\x00\x00\xff"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: Hex = field(XmpHex(size=1))
        custom_field2: Hex = field(XmpHex(size=3))

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, str)
    assert obj.custom_field == "00"
    assert obj.custom_field == Hex("00")
    assert obj.custom_field2 == "0000ff"
    assert obj.custom_field2 == Hex("0000ff")


def test_hex_string() -> None:
    data = b'\x04\xf4\xbc\x9cs\xd1\x04\xf4\xbc\x9cs\xd0\xff\xff\x00\x00'

    class SetDataAttr(ResponseBodyStruct):
        hex_data: Hex = field(XmpHex())

    obj = SetDataAttr(data)
    assert obj.hex_data[:-4] == Hex("04F4BC9C73D104F4BC9C73D0FFFF").lower()


def test_ip_v4() -> None:
    data = b'\x7f\x00\x00\x01'

    class GetDataAttr(ResponseBodyStruct):
        custom_field: IPv4Address = field(XmpIPv4Address())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, IPv4Address)
    assert obj.custom_field == IPv4Address("127.0.0.1")
    assert obj.custom_field != "127.0.0.1"


def test_ip_v6() -> None:
    data = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab'

    class GetDataAttr(ResponseBodyStruct):
        custom_field: IPv6Address = field(XmpIPv6Address())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, IPv6Address)
    assert obj.custom_field == IPv6Address("::ab")
    assert obj.custom_field != "::ab"


def test_mac() -> None:
    data = b'\xf4\xd4\x88\x67\xc5\xda'

    class GetDataAttr(ResponseBodyStruct):
        custom_field: Hex = field(XmpMacAddress())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, str)
    assert obj.custom_field == "f4d48867c5da"
    assert obj.custom_field != "f4:d4:88:67:c5:da"


def test_string() -> None:
    data_1 = b"Odin-10G-1S-6P[b]\x00\x00\n"
    data_2 = b"Odin-10G-1S-6P[b]\x00\x00\x0c"
    data_3 = b""

    class GetDataAttr(ResponseBodyStruct):
        custom_field: str = field(XmpStr())

    obj1 = GetDataAttr(data_1)
    obj2 = GetDataAttr(data_2)
    obj3 = GetDataAttr(data_3)

    assert isinstance(obj1.custom_field, str)
    assert isinstance(obj2.custom_field, str)
    assert isinstance(obj3.custom_field, str)
    assert obj1.custom_field == "Odin-10G-1S-6P[b]"
    assert obj2.custom_field == "Odin-10G-1S-6P[b]"
    assert obj3.custom_field == ""


def test_unlimited_list() -> None:
    data = b'\x00\x02\x00\x06\x00\x06\x00\x06'

    class GetDataAttr(ResponseBodyStruct):
        port_counts: List[int] = field(XmpSequence(types_chunk=[XmpByte()]))

    obj = GetDataAttr(data)
    assert hasattr(obj, "port_counts")
    assert obj.port_counts == [0, 2, 0, 6, 0, 6, 0, 6]


def test_limited_list() -> None:
    data = b'\x00\x02\x00\x06\x00\x06\x00\x06'
    MAX_LEN = 2

    class GetDataAttr(ResponseBodyStruct):
        custom_field: List[int] = field(XmpSequence(types_chunk=[XmpByte()], length=MAX_LEN))

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, list)
    assert len(obj.custom_field) == MAX_LEN
    assert obj.custom_field == [0, 2]

# endregion


# region TypeConversion

class UsersEnum(Enum):
    OPTION_1 = 1
    OPTION_2 = 2
    OPTION_3 = 3
    OPTION_4 = 5


def test_enum_global_def() -> None:
    data = b"\x02\x05"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: UsersEnum = field(XmpByte())
        custom_field2: UsersEnum = field(XmpByte())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, UsersEnum)
    assert isinstance(obj.custom_field2, UsersEnum)
    assert obj.custom_field == UsersEnum.OPTION_2
    assert obj.custom_field2 == UsersEnum.OPTION_4


def test_enum_local_def() -> None:
    """
    At the Moment the defenition of the enum under the local namespace will faile
    """
    data = b"\x02\x05"

    class UsersLocalEnum(Enum):
        OPTION_1 = 1
        OPTION_2 = 2
        OPTION_3 = 3
        OPTION_4 = 5

    class GetDataAttr(ResponseBodyStruct):
        custom_field: UsersLocalEnum = field(XmpByte())
        custom_field2: UsersLocalEnum = field(XmpByte())

    obj = GetDataAttr(data)
    try:
        assert isinstance(obj.custom_field, UsersLocalEnum)
        assert isinstance(obj.custom_field2, UsersLocalEnum)
        assert obj.custom_field == UsersLocalEnum.OPTION_2
        assert obj.custom_field2 == UsersLocalEnum.OPTION_4
    except TypeError as e:
        assert str(e) == "'ForwardRef' object is not callable"
    else:
        assert False, "Magicly it start to works"


def test_bool() -> None:
    data = b"\x01\x00"

    class GetDataAttr(ResponseBodyStruct):
        custom_field: bool = field(XmpByte())
        custom_field2: bool = field(XmpByte())

    obj = GetDataAttr(data)
    assert isinstance(obj.custom_field, bool)
    assert isinstance(obj.custom_field2, bool)
    assert obj.custom_field is True
    assert obj.custom_field2 is False


class OnOff(IntEnum):
    OFF = 0
    ON = 1


@dataclass
class ArpChunk:
    ipv4_address: IPv4Address
    prefix: int
    patched_mac: OnOff
    mac_address: Hex


def test_chunk_type() -> None:
    data = b'\x7f\x00\x00\x01\x02\x05\x01\xf4\xd4\x88\x67\xc5\xda\x7f\x00\x00\x03\x02\x05\x00\xf4\xd4\x88\x55\xc5\xda'

    class GetDataAttr(ResponseBodyStruct):
        custom_field: List[ArpChunk] = field(XmpSequence(types_chunk=[XmpIPv4Address(), XmpShort(), XmpByte(), XmpMacAddress()]))
    obj = GetDataAttr(data)
    assert len(obj.custom_field) == 2
    assert obj.custom_field[0].ipv4_address == IPv4Address("127.0.0.1")
    assert obj.custom_field[0].prefix == 517
    assert obj.custom_field[0].patched_mac == OnOff.ON
    assert obj.custom_field[0].mac_address == "f4d48867c5da"

# endregion


# region TypeCombinations

# endregion



from typing import Set


class WrongModuleError(Exception):
    def __init__(self, module_revision: str, allowed_revisions: Set[str]) -> None:
        self.module_revision = module_revision
        self.allowed_revisions = allowed_revisions
        self.msg = (
            f"Module of revision <{self.module_revision}> can not be assign to the current tester,\n"
            f"Permitted only modules of next revisions {allowed_revisions}"
        )
        super().__init__(self.msg)



from __future__ import annotations
from xoa_driver.enums import MediaConfigurationType
from xoa_driver.functions.mgmt import (
    get_module_supported_media,
    set_module_media_config,
)

import timeit
import pstats
import cProfile

import asyncio
import os
import sys
import logging
from typing import Coroutine, List, Any

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from loguru import logger  # noqa: E402
from xoa_driver.testers import L23Tester  # noqa: E402
from xoa_driver.internals.utils.managers.exceptions import NoSuchPortError  # noqa: E402


async def test_using_a_outdated_port():
    # Module 7 of destination chassis is supposed to have two modes: BASE-T1(7 ports), BASE-T1S(2 ports), switch to BASE-T1 to start this test.
    t = await L23Tester("192.168.1.198", "Ron", enable_logging=False)

    m1 = t.modules.obtain(7)
    p1 = m1.ports.obtain(3)
    print(await p1.interface.get())
    print('success1')
    await set_module_media_config(m1, MediaConfigurationType.BASE_T1S)
    try:
        print(await p1.interface.get())
        print('success2')
    except ConnectionRefusedError as e:
        print(e)
    try:
        p2 = m1.ports.obtain(3)
        print(await p2.interface.get())
        print('success3')
    except NoSuchPortError as e:
        print(e)


async def test_getting_a_removed_port():
    # Module 7 of destination chassis is supposed to have two modes: BASE-T1(7 ports), BASE-T1S(2 ports), switch to BASE-T1 to start this test.
    t = await L23Tester("192.168.1.198", "Ron", enable_logging=False)
    m1 = t.modules.obtain(7)
    print(len(m1.ports))

    p1 = m1.ports.obtain(1)
    print('success1', await p1.interface.get())

    await set_module_media_config(m1, MediaConfigurationType.BASE_T1S)

    print(len(m1.ports))
    try:
        print('success2', await p1.interface.get())
    except ConnectionRefusedError as e:
        print(e)
    # This will raise Exceptions.

    p2 = m1.ports.obtain(1)
    print('success3', await p2.interface.get())
    # This line should do fine.


async def test_getting_stream_under_a_removed_port():
    # Module 7 of destination chassis is supposed to have two modes: BASE-T1(7 ports), BASE-T1S(2 ports), switch to BASE-T1 to start this test.
    t = await L23Tester("192.168.1.198", "Ron", enable_logging=False)
    m1 = t.modules.obtain(7)
    p1 = m1.ports.obtain(3)
    await p1.reservation.set_reserve()
    s = await p1.streams.create()
    b = s.comment
    await set_module_media_config(m1, MediaConfigurationType.BASE_T1S)

    print(await b.get())


def run(method: Coroutine) -> None:
    import platform

    if platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.run(method)


if __name__ == "__main__":
    run(test_getting_stream_under_a_removed_port())



from dataclasses import dataclass
from typing import (
    Callable,
    Type,
    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core.transporter.protocol.struct_response import Response
    from xoa_driver.internals.core.transporter._typings import XoaCommandType
    from xoa_driver.internals.core import interfaces as itf


@dataclass
class Update:
    inst: Type
    property_name: str
    response_key: str
    condition: Callable[["Response"], bool] = lambda _: True

    # keep it Async just for consistent interface of event_observer
    async def __call__(self, response: "Response") -> None:
        if not self.condition(response):
            return None
        v = getattr(response.values, self.response_key)
        setattr(self.inst, self.property_name, v)


def on_event(self, evt: "XoaCommandType", callback: "itf.CallbackType") -> None:
    async def _f(response: "Response", *args, **kwargs) -> None:
        if not self._check_identity(response):
            return None
        await callback(self, response.values, *args, **kwargs)
    self._conn.subscribe(evt, _f)



import asyncio
import os
import sys
import pytest

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))  # IMPORTANT: Suppose to stay before driver import

from xoa_driver.functions.mgmt import reserve_port, reserve_module, reserve_tester
from xoa_driver.testers import L23Tester
from typing import Coroutine



from xoa_driver.functions.config_cli_convert import (
    read_commands_from_file,
    read_commands_from_string,
    upload_port_config_from_string,
    upload_port_config_from_file,
    upload_module_config_from_string,
    upload_module_config_from_file,
    upload_tester_config_from_string,
    upload_tester_config_from_file,
)  # noqa: E402


@pytest.mark.asyncio
async def test_read_commands_from_file() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_all_commands.txt")
    for c in read_commands_from_file(path):
        print(c)
        print(c.as_request())


@pytest.mark.asyncio
async def test_read_commands_from_string() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_all_commands.txt")
    with open(path, "r") as f:
        content = f.read()
        for c in read_commands_from_string(content):
            print(c)
            print(c.as_request())


@pytest.mark.asyncio
async def test_upload_port_config_from_file() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_port_config.txt")
    module = (await L23Tester("192.168.1.198", "Happy", enable_logging=True)).modules.obtain(1)
    port = module.ports.obtain(0)
    await reserve_port(port, True)
    await upload_port_config_from_file(port, path)


@pytest.mark.asyncio
async def test_upload_port_config_from_string() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_port_config.txt")
    with open(path, "r") as f:
        content = f.read()
        module = (await L23Tester("192.168.1.198", "Happy", enable_logging=True)).modules.obtain(1)
        port = module.ports.obtain(0)
        await reserve_port(port, True)
        await upload_port_config_from_string(port, content)


@pytest.mark.asyncio
async def test_upload_module_config_from_file() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_module_config.txt")
    module = (await L23Tester("192.168.1.198", "Happy", enable_logging=True)).modules.obtain(1)
    await reserve_module(module, True)
    await upload_module_config_from_file(module, path)


@pytest.mark.asyncio
async def test_upload_module_config_from_string() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_module_config.txt")
    with open(path, "r") as f:
        content = f.read()
        module = (await L23Tester("192.168.1.198", "Happy", enable_logging=True)).modules.obtain(1)
        await reserve_module(module, True)
        await upload_module_config_from_string(module, content)


@pytest.mark.asyncio
async def test_upload_tester_config_from_string() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_tester_config.txt")
    with open(path, "r") as f:
        content = f.read()
        tester = await L23Tester("192.168.1.198", "Happy", enable_logging=True)
        await reserve_tester(tester, True)
        await upload_tester_config_from_string(tester, content)


@pytest.mark.asyncio
async def test_upload_tester_config_from_file() -> None:
    path = os.path.join(os.path.dirname(__file__), "test_clis", "test_tester_config.txt")
    tester = await L23Tester("192.168.1.198", "Happy", enable_logging=True)
    await reserve_tester(tester, True)
    await upload_tester_config_from_file(tester, path)


def run(method: Coroutine) -> None:
    import platform

    if platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.run(method)


if __name__ == "__main__":
    run(test_upload_module_config_from_file())
    # run(test_upload_module_config_from_string())
    # run(test_upload_port_config_from_file())
    # run(test_upload_port_config_from_string())
    # run(test_upload_tester_config_from_string())
    # run(test_upload_tester_config_from_file())



from __future__ import annotations
from typing import (
    Any,
    Awaitable,
    TypeVar,
    # Generic,
    Generator
)
from dataclasses import dataclass

from . import interfaces
from .transporter.protocol.struct_request import Request


AwaitableDataType = TypeVar("AwaitableDataType", covariant=True)


@dataclass(frozen=True)
class Token(Awaitable[AwaitableDataType]):
    """
    A wrapper of connection and request.
    Which can be used for await an single command or organized in to the ordered
    sequence of the commands which will be send to the server in single request.
    """
    __slots__ = ('connection', 'request',)

    connection: interfaces.IConnection
    request: Request

    def __await__(self) -> Generator[Any, None, AwaitableDataType]:
        return self.__ask().__await__()

    async def __ask(self) -> AwaitableDataType:
        (
            data,
            fut
        ) = await self.connection.prepare_data(self.request)
        self.connection.send(data)
        try:
            result = await fut
        except Exception as e:
            raise e
        else:
            return result



from __future__ import annotations
from dataclasses import dataclass
import typing

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpInt,
    XmpLong,
    XmpSequence,
    XmpStr
)
from .enums import (
    OnOff,
    LatencyTypeCustomDist,
)


@register_command
@dataclass
class PEC_INDICES:
    """
    The full list of which custom distributions which are defined for a port. These
    are the custom id values that are used for assigning the custom distributions to
    an impairment. Setting the value of this command creates a new custom
    distribution (default values) for each value that is not already in use, and
    deletes each custom distribution that is not mentioned in the list. The same can
    be accomplished one-custom-distribution-at-a-time using the PEC_VAL and
    PEC_DELETE commands.

    .. note::

        Custom distributions which are currently defined are not affected when mentioned in a PEC_INDICES set command.
        Custom distributions which are currently assigned to an impairment cannot be deleted and any attempt of deleting
        such a custom distribution using either PEC_DELETE` or PEC_INDICES` will result in an error.

    """

    code: typing.ClassVar[int] = 1610
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        indexations: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))
        """list of integers, a list of the indices to the custom distributions which are currently defined on that port, max 40 elements."""

    class SetDataAttr(RequestBodyStruct):
        indexations: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))
        """list of integers, a list of the indices to the custom distributions which are currently defined on that port, max 40 elements."""

    def get(self) -> Token[GetDataAttr]:
        """Get a list of the indices to the custom distributions which are currently defined on that port.

        :return: a list of the indices to the custom distributions which are currently defined on that port
        :rtype: PEC_INDICES.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, indexations: typing.List[int]) -> Token[None]:
        """Set a list of indices to create new custom distributions.

        :param indexations: a list of indices to create new custom distributions
        :type indexations: typing.List[int]
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indexations=indexations))


@register_command
@dataclass
class PEC_VAL:
    """
    Definition of custom distribution. Custom distributions can be defined for
    latency with 1024 entries and for non-latency impairments with 512 entries. Each
    port will maintain a list of defined custom distributions, identified by an
    CUST_ID. (Range: 1 - 40).
    """

    code: typing.ClassVar[int] = 1680
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _custom_distribution_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        linear: OnOff = field(XmpByte())
        """coded byte, defines the way the FPGA RAM content is played out."""
        symmetric: OnOff = field(XmpByte())
        """coded byte, reserved for future use, must be set to OFF."""
        entry_count: int = field(XmpInt())
        """integer, defines the number of entries in "dataX" (allowed value: 512,1024). NOTE: For Latency, 1024 entries are used, and for rest, 512 entries are used)"""
        data_x: typing.List[int] = field(XmpSequence(types_chunk=[XmpLong()]))
        """array of long integers, array size="num_entries", holds values to be filled in the RAM memory."""

    class SetDataAttr(RequestBodyStruct):
        linear: OnOff = field(XmpByte())
        """coded byte, defines the way the FPGA RAM content is played out."""
        symmetric: OnOff = field(XmpByte())
        """coded byte, reserved for future use, must be set to OFF."""
        entry_count: int = field(XmpInt())
        """integer, defines the number of entries in "dataX" (allowed value: 512,1024). NOTE: For Latency, 1024 entries are used, and for rest, 512 entries are used)"""
        data_x: typing.List[int] = field(XmpSequence(types_chunk=[XmpLong()]))
        """array of long integers, array size="num_entries", holds values to be filled in the RAM memory."""

    def get(self) -> Token[GetDataAttr]:
        """Get the definition of custom distribution.

        :return: definition of custom distribution
        :rtype: PEC_VAL.GetDataAttr.
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._custom_distribution_xindex]))

    def set(self, linear: OnOff, symmetric: OnOff, entry_count: int, data_x: typing.List[int]) -> Token[None]:
        """Set the definition of custom distribution. Custom distributions can be defined for
        latency with 1024 entries and for non-latency impairments with 512 entries. Each
        port will maintain a list of defined custom distributions, identified by an
        CUST_ID. (Range: 1 - 40).

        :param linear: defines the way the FPGA RAM content is played out
        :type linear: OnOff
        :param symmetric: reserved for future use, must be set to 0.
        :type symmetric: OnOff
        :param entry_count: defines the number of entries in "data_x" (allowed value: 512,1024). For Latency, 1024 entries are used, and for rest, 512 entries are used.
        :type entry_count: int
        :param data_x: array size equals to "entry_count", holds values to be filled in the RAM memory.
        :type data_x: typing.List[int]
        """

        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._custom_distribution_xindex],
                linear=linear,
                symmetric=symmetric,
                entry_count=entry_count,
                data_x=data_x
            )
        )


@register_command
@dataclass
class PEC_COMMENT:
    """
    Defines the user-defined description string of a custom distribution.
    """

    code: typing.ClassVar[int] = 1681
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _custom_distribution_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        comment: str = field(XmpStr())
        """string, the user-specified comment/description for the custom distribution."""

    class SetDataAttr(RequestBodyStruct):
        comment: str = field(XmpStr())
        """string, the user-specified comment/description for the custom distribution."""

    def get(self) -> Token[GetDataAttr]:
        """Get the user-defined description string of a custom distribution.

        :return: the user-specified comment/description for the custom distribution.
        :rtype: PEC_COMMENT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._custom_distribution_xindex]))

    def set(self, comment: str) -> Token[None]:
        """Set the user-defined description string of a custom distribution.

        :param comment: the user-specified comment/description for the custom distribution.
        :type comment: str
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._custom_distribution_xindex], comment=comment))


@register_command
@dataclass
class PEC_DELETE:
    """
    Deletes the custom distribution definition.

    .. note::

        Once a customer has defined a customer distribution using PEC_VAL, it is defined until it is explicitly deleted.
        Only customer distributions which are not referenced by any impairments, can be deleted.

    """

    code: typing.ClassVar[int] = 1682
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _custom_distribution_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Deletes the custom distribution definition.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._custom_distribution_xindex]))


@register_command
@dataclass
class PEC_DISTTYPE:
    """
    Retrieves if a custom distribution is defined for latency or non-latency.

    .. note::

        Using PEC_DISTTYPE as set has no effect. The distribution type is determined upon custom distribution creation and cannot be modified later.
        However, it is legal to issue the PEC_DISTTYPE set command with no effect.

    """

    code: typing.ClassVar[int] = 1683
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _custom_distribution_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        latency_type: LatencyTypeCustomDist = field(XmpByte())
        """byte, 0 indicates interpacket distribution, 1 indicates latency distribution."""

    def get(self) -> Token[GetDataAttr]:
        """Get the latency type of a custom distribution.

        :return: latency type of a custom distribution
        :rtype: PEC_DISTTYPE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._custom_distribution_xindex]))



import asyncio
from typing import (
    Optional,
    Final,
)
from xoa_driver.internals.core import funcs
from xoa_driver.internals.commands import enums
from xoa_driver.internals.commands import (
    C_RESERVATION,
    C_CAPABILITIES,
    C_RESERVEDBY,
)
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.exceptions.testers import UnsupportedFirmwareError

MIN_SUPPORTED_VERSION = 446.5


class TesterLocalState:
    """Tester local state

    :param host: tester's address/hostname
    :type host: str
    :param port: the port number for connection establishment
    :type port: int
    """
    __slots__ = (
        "host",
        "port",
        "model",
        "capabilities",
        "serial_number",
        "version_major",
        "driver_version",
        "reservation",
        "reserved_by",
    )

    def __init__(self, host: str, port: int) -> None:
        self.host: Final[str] = host
        self.port: Final[int] = port
        self.model: str = ""
        self.capabilities: Optional["C_CAPABILITIES.GetDataAttr"] = None
        self.serial_number: int = 0
        self.version_major: int = 0
        self.driver_version: int = 0
        self.reservation: enums.ReservedStatus = enums.ReservedStatus.RELEASED
        self.reserved_by: str = ""

    async def initiate(self, tester) -> None:
        (
            capabilities_resp,
            model_res,
            v_major_res,
            serial_res,
            reserved_by_res,
            reservation_resp,
        ) = await funcs.apply(
            tester.capabilities.get(),
            tester.model.get(),
            tester.version_no.get(),
            tester.serial_no.get(),
            tester.reserved_by.get(),
            tester.reservation.get(),
        )
        self.reserved_by = reserved_by_res.username
        self.model = model_res.model
        self.driver_version = v_major_res.pci_driver_version
        self.version_major = v_major_res.chassis_major_version
        self.serial_number = serial_res.serial_number
        self.reservation = reservation_resp.operation
        self.capabilities = capabilities_resp

    def register_subscriptions(self, tester) -> None:
        tester._conn.subscribe(C_RESERVEDBY, utils.Update(self, "reserved_by", "username"))
        tester._conn.subscribe(C_RESERVATION, utils.Update(self, "reservation", "operation"))


class GenuineTesterLocalState(TesterLocalState):
    """Genuine Tester Local State

    :param host: tester's address/hostname
    :type host: str
    :param port: the port number for connection establishment
    :type port: int
    """
    __slots__ = ("build_string", "version_minor")

    def __init__(self, host: str, port: int) -> None:
        super().__init__(host, port)
        self.build_string: str = ""
        self.version_minor: int = 0

    async def initiate(self, tester) -> None:
        bs, v_minor, _ = await asyncio.gather(
            tester.build_string.get(),
            tester.version_no_minor.get(),
            super().initiate(tester)
        )
        self.build_string = bs.build_string
        self.version_minor = v_minor.chassis_minor_version
        current_version = float(f"{self.version_major}.{self.version_minor}")
        if current_version < MIN_SUPPORTED_VERSION:
            raise UnsupportedFirmwareError(current_version)



from __future__ import annotations
from asyncio import Future
from typing import Generic
from typing_extensions import Self
from ._typings import HeaderType


class StreamReader(Generic[HeaderType]):
    __slots__ = ("_buffer", "_eof", "_waiter", "__header_struct")

    def __init__(self, header_struct: type[HeaderType]) -> None:
        self._buffer = bytearray()
        self._eof = False  # Whether we're done.
        self._waiter = None  # A future used by _wait_for_data()
        self.__header_struct = header_struct

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> tuple[HeaderType, bytes]:
        try:
            val = await self.read_pkt()
        except EOFError:
            raise StopAsyncIteration
        else:
            return val

    def _wakeup_waiter(self) -> None:
        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_result(None)

    async def _wait_for_data(self, func_name: str) -> None:
        """Wait until feed_data() or feed_eof() is called.
        If stream was paused, automatically resume it.
        """
        if self._waiter is not None:
            raise RuntimeError(
                f'{func_name}() called while another coroutine is '
                f'already waiting for incoming data'
            )

        assert not self._eof, '_wait_for_data after EOF'

        self._waiter = Future()
        try:
            await self._waiter
        finally:
            self._waiter = None

    def feed_eof(self) -> None:
        self._eof = True
        self._wakeup_waiter()

    def at_eof(self) -> bool:
        """Return True if the buffer is empty and 'feed_eof' was called."""
        return self._eof and not self._buffer

    def feed_data(self, data: bytes) -> None:
        assert not self._eof, 'feed_data after feed_eof'

        if not data:
            return None

        self._buffer.extend(data)
        self._wakeup_waiter()

    async def readexactly(self, n: int) -> bytes:
        if n < 0:
            raise ValueError('readexactly size can not be less than zero')

        if n == 0:
            return b''

        while len(self._buffer) < n:
            if self._eof:
                incomplete = bytes(self._buffer)
                self._buffer.clear()
                raise EOFError(incomplete, n)

            await self._wait_for_data('readexactly')

        if len(self._buffer) == n:
            data = bytes(self._buffer)
            self._buffer.clear()
        else:
            data = bytes(self._buffer[:n])
            del self._buffer[:n]
        return data

    async def read_pkt(self) -> tuple[HeaderType, bytes]:
        h_buff = await self.readexactly(self.__header_struct.size)
        header = self.__header_struct.from_bytes(h_buff)
        if not header:
            raise ValueError("Invalid Header")
        body_bytes = await self.readexactly(header.body_size)
        return (header, body_bytes)



import typing
from xoa_driver import ports
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils.managers import ports_manager as pm

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23


@typing.final
@revisions.register_valkyrie_module(rev="Odin-5G-4S-6P-CU")
class MOdin5G4S6PCU(ModuleL23):
    """Test module Odin-5G-4S-6P-CU"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin5G4S6PCU] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin5G4S6PCU,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-5G-4S-6P-CU"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-5S-6P-CU")
class MOdin10G5S6PCU(ModuleL23):
    """Test module Odin-10G-5S-6P-CU"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G5S6PCU] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G5S6PCU,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-5S-6P-CU"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-5S-6P-CU[b]")
class MOdin10G5S6PCU_b(ModuleL23):
    """Test module Odin-10G-5S-6P-CU[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G5S6PCU_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G5S6PCU_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-5S-6P-CU[b]"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-3S-6P-CU")
class MOdin10G3S6PCU(ModuleL23):
    """Test module Odin-10G-3S-6P-CU"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G3S6PCU] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G3S6PCU,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-3S-6P-CU"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-3S-2P-CU")
class MOdin10G3S2PCU(ModuleL23):
    """Test module Odin-10G-3S-2P-CU"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G3S2PCU] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G3S2PCU,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-3S-2P-CU"""



from typing import TYPE_CHECKING
from xoa_driver.internals.commands import (
    P_AUTONEGSELECTION,  # questionable which ports are electrical
    P_LPENABLE,
    P_LPTXMODE,
    P_LPSTATUS,
    P_LPPARTNERAUTONEG,
    P_LPSNRMARGIN,
    P_LPRXPOWER,
    P_LPSUPPORT,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from .bases.port_l23_genuine import BasePortL23Genuine


class LowPowerMode:
    """L23 port low power mode."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.enable = P_LPENABLE(conn, module_id, port_id)
        """Energy Efficiency Ethernet.
        
        :type: P_LPENABLE
        """

        self.mode = P_LPTXMODE(conn, module_id, port_id)
        """Low power TX mode.
        
        :type: P_LPTXMODE
        """
        self.status = P_LPSTATUS(conn, module_id, port_id)
        """Low power status.
        
        :type: P_LPSTATUS
        """

        self.partner_capabilities = P_LPPARTNERAUTONEG(conn, module_id, port_id)
        """EEE capabilities advertised during auto-negotiation by the far side.
        
        :type: P_LPPARTNERAUTONEG
        """

        self.snr_margin = P_LPSNRMARGIN(conn, module_id, port_id)
        """SNR margin on the four link channels by the PHY.
        
        :type: P_LPSNRMARGIN
        """

        self.rx_power = P_LPRXPOWER(conn, module_id, port_id)
        """RX power recorded during training for the four channels.
        
        :type: P_LPRXPOWER
        """

        self.capabilities = P_LPSUPPORT(conn, module_id, port_id)
        """EEE capabilities of the port.
        
        :type: P_LPSUPPORT
        """


class FamilyE(BasePortL23Genuine):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.autoneg_selection = P_AUTONEGSELECTION(conn, module_id, port_id)
        """L23 port's auto-negotiation selection.
        
        :type: P_AUTONEGSELECTION
        """

        self.eee = LowPowerMode(conn, module_id, port_id)
        """L23 port Low Power mode settings.
        
        :type: LowPowerMode
        """


class POdin5G4S6PCU(FamilyE):
    """L23 port on Odin-5G-4S-6P-CU module.
    """
    ...


class POdin10G5S6PCU(FamilyE):
    """L23 port on Odin-10G-5S-6P-CU module.
    """
    ...


class POdin10G5S6PCU_b(FamilyE):
    """L23 port on Odin-10G-5S-6P-CU[b] module.
    """
    ...


class POdin10G3S6PCU(FamilyE):
    """L23 port on Odin-10G-3S-6P-CU module.
    """
    ...


class POdin10G3S2PCU(FamilyE):
    """L23 port on Odin-10G-3S-2P-CU module.
    """
    ...



from __future__ import annotations
from typing import Any

from xoa_driver.ports import GenericAnyPort
from xoa_driver.internals.core import interfaces as itf
from xoa_driver.lli import commands
from xoa_driver import enums


def get_ctx(port: GenericAnyPort) -> tuple["itf.IConnection", int, int]:
    return (port._conn, *port.kind)


def dictionize_autoneg_status(
    loopback: commands.PL1_CFG_TMP.GetDataAttr,
    auto_neg_info: commands.PL1_AUTONEGINFO.GetDataAttr,
    status: commands.PL1_AUTONEG_STATUS.GetDataAttr,
) -> dict[str, Any]:
    _is_enabled = True if status.mode == enums.AutoNegMode.ANEG_ON else False
    _ta_hcd_status = status.tech_ability_hcd_status
    if _ta_hcd_status == enums.FreyaTechAbilityHCDStatus.FAILED:
        _ta_hcd_value = "N/A"
        _fec_result_value = "N/A"
    else:
        _ta_hcd_value = status.tech_ability_hcd_value.name
        _fec_result_value = status.fec_mode_result.name
    return {
        "is_enabled": _is_enabled,
        "loopback": "allowed" if loopback.values[0] else "not allowed",
        "hcd": _ta_hcd_value,
        "fec_result": _fec_result_value,
        "duration": auto_neg_info.duration_us,
        "successes": auto_neg_info.negotiation_success_count,
        "timeouts": auto_neg_info.negotiation_timeout_count,
        "loss_of_sync": auto_neg_info.negotiation_loss_of_sync_count,
        "fec_negotiation_fails": auto_neg_info.negotiation_fec_fail_count,
        "hcd_negotiation_fails": auto_neg_info.negotiation_hcd_fail_count,
        "link_codewords": {
            "tx": auto_neg_info.tx_link_codeword_count,
            "rx": auto_neg_info.rx_link_codeword_count,
        },
        "next_page_messages": {
            "tx": auto_neg_info.tx_next_page_message_count,
            "rx": auto_neg_info.rx_next_page_message_count,
        },
        "unformatted_pages": {
            "tx": auto_neg_info.tx_next_page_unformatted_count,
            "rx": auto_neg_info.rx_next_page_unformatted_count,
        },
    }


def _decode_ic(key: int) -> str:
    dic = {
        0: "INDV",
        1: "Preset 4",
        2: "Preset 1",
        3: "Preset 5",
        4: "Preset 2",
        6: "Preset 3",
    }
    return dic.get(key, "Reserved")


def _link_info_part(info: commands.PL1_LINKTRAININFO.GetDataAttr, index: str, part: str) -> dict[str, str]:
    tx_attr = f"{index}_tx_{part}"
    rx_attr = f"{index}_rx_{part}"
    return {
        "rx": getattr(info, rx_attr),
        "tx": getattr(info, tx_attr),
    }


def _link_info_all(info: commands.PL1_LINKTRAININFO.GetDataAttr, index: str) -> dict[str, Any]:
    return {
        "current_level": getattr(info, f"{index}_current_level"),
        "+req": _link_info_part(info, index, "increment_req_count"),
        "-req": _link_info_part(info, index, "decrement_req_count"),
        "coeff_and_eq_limit_reached": _link_info_part(
            info, index, "coeff_eq_limit_reached_count"
        ),
        "eq_limit_reached": _link_info_part(info, index, "eq_limit_reached_count"),
        "coeff_not_supported": _link_info_part(
            info, index, "coeff_not_supported_count"
        ),
        "coeff_at_limit": _link_info_part(info, index, "coeff_at_limit_count"),
    }


def dictionize_lt_status(
    status: commands.PL1_LINKTRAIN_STATUS.GetDataAttr,
    info: commands.PL1_LINKTRAININFO.GetDataAttr,
    ltconf: commands.PL1_LINKTRAIN_CONFIG.GetDataAttr,
    cfg: commands.PL1_CFG_TMP.GetDataAttr,
    ber: float,
    total_bit_count: float,
    total_error_bit_count: float,
) -> dict[str, Any]:
    _is_enabled = True if status.mode == enums.LinkTrainingStatusMode.ENABLED else False
    _is_traind = True if status.status == enums.LinkTrainingStatus.TRAINED else False
    _oos_preset = "Existing" if ltconf.oos_preset == enums.FreyaOutOfSyncPreset.CURRENT else "IEEE"
    ber_str = '{:.2e}'.format(ber)
    return {
        "is_enabled": _is_enabled,
        "is_trained": _is_traind,
        # "failure": enums.LinkTrainingFailureType(status.failure).name.lower(),
        "failure": status.failure.name.lower(),
        "oos_preset": _oos_preset,
        "init_modulation": enums.LinkTrainEncoding(cfg.values[0]).name.lower(),
        "total_bits": total_bit_count,
        "total_errored_bits": total_error_bit_count,
        "ber": ber_str,
        "duration": info.duration_us,
        "lock_lost": info.lock_lost_count,
        # "frame_lock": enums.LinkTrainFrameLock(info.frame_lock).name.lower(),
        "frame_lock": info.frame_lock.name.lower(),
        # "remote_frame_lock": enums.LinkTrainFrameLock(
        #     info.remote_frame_lock
        # ).name.lower(),
        "remote_frame_lock": info.remote_frame_lock.name.lower(),
        "frame_errors": info.num_frame_errors,
        "overrun_errors": info.num_overruns,
        "last_ic_received": _decode_ic(info.last_ic_received),
        "last_ic_sent": _decode_ic(info.last_ic_sent),
        "c(-3)": _link_info_all(info, "pre3"),
        "c(-2)": _link_info_all(info, "pre2"),
        "c(-1)": _link_info_all(info, "pre1"),
        "c(0)": _link_info_all(info, "main"),
        "c(1)": _link_info_all(info, "post1"),
    }


def dictionize_txtap_get(r: commands.PP_PHYTXEQ.GetDataAttr) -> dict[str, int]:
    return {
        "c(-3)": r.pre3_post2,
        "c(-2)": r.pre2,
        "c(-1)": r.pre,
        "c(0)": r.main,
        "c(1)": r.post,
    }


def dictionize_anlt_status(
    link_recovery: commands.PL1_CFG_TMP.GetDataAttr,
    anlt_op: commands.PL1_ANLT.GetDataAttr,
    linktrain_cfg: commands.PL1_LINKTRAIN_CONFIG.GetDataAttr,
    capabilities: commands.P_CAPABILITIES.GetDataAttr,
    allow_loopback: commands.PL1_CFG_TMP.GetDataAttr,
) -> dict[str, Any]:
    return {
        "autoneg_mode": anlt_op.an_mode.name.lower(),
        "link_training_mode": anlt_op.lt_mode.name.lower(),
        "link_training_timeout": "enable" if linktrain_cfg.timeout_mode == enums.TimeoutMode.DEFAULT else "disable",
        "restart_link_down": "on" if link_recovery.values[0] == 1 or link_recovery.values[0] == 3 else "off",
        "restart_lt_fail": "on" if link_recovery.values[0] == 2 or link_recovery.values[0] == 3 else "off",
        "serdes_count": capabilities.serdes_count,
        "autoneg_allow_loopback": allow_loopback.values,
        "link_training_preset0": enums.FreyaOutOfSyncPreset(linktrain_cfg.oos_preset).name.lower(),
    }


def dictionize_lt_im_status(
    capabilities: commands.P_CAPABILITIES.GetDataAttr,
    initial_mods: dict[str, str]
) -> dict[str, Any]:
    return {
        "serdes_count": capabilities.serdes_count,
        "initial_mods": initial_mods,
    }


def dictionize_lt_algorithm_status(
    capabilities: commands.P_CAPABILITIES.GetDataAttr,
    algorithms: dict[str, str]
) -> dict[str, Any]:
    return {
        "serdes_count": capabilities.serdes_count,
        "algorithms": algorithms
    }


MODULE_EOL_INFO: dict[str, str] = {
    "01": "2014-04-01",
    "02": "2024-09-01",
    "03": "2016-03-01",
    "09": "2022-01-01",
    "17": "2023-01-01",
    "18": "2023-01-01",
    "20": "2024-11-01",
    "22": "2018-11-01",
    "24": "2024-11-01",
    "26": "2023-06-01",
    "27": "2025-10-01",
    "30": "2024-01-01",
    "31": "2021-09-01",
    "32": "2024-04-01",
    "34": "2024-08-01",
    "36": "2024-04-01",
    "40": "2023-03-01",
    "50": "2022-02-01",
    "51": "2023-08-01",
    "54": "2023-01-01",
    "55": "2024-01-01",
    "60": "2025-10-01",
    "66": "2025-01-31",
    "90": "2025-10-01",
    "91": "2025-10-01",
    "93": "2025-10-01",
    "94": "2025-10-01",
    "97": "2025-10-01",
}



from __future__ import annotations
import asyncio
import typing as t
from functools import partial
from xoa_driver import enums
from xoa_driver.ports import GenericL23Port
from xoa_driver.lli import commands
from xoa_driver.misc import Hex
from dataclasses import dataclass
from enum import IntEnum
from .tools import get_ctx


class AnLtD(IntEnum):
    PMD_CONFIG_REGISTER = 0x02
    AN_TX_CONFIG_REGISTER = 0x10
    AN_RX_CONFIG_REGISTER = 0x18
    AN_RX_STATUS_REGISTER = 0x19
    AN_TX_PAGE_0_REGISTER = 0x14
    AN_TX_PAGE_1_REGISTER = 0x15
    AN_RX_DME_MV_RANGE = 0x1A
    AN_RX_DME_BIT_RANGE = 0x1B
    AN_RX_PAGE_0_REGISTER = 0x1C
    AN_RX_PAGE_1_REGISTER = 0x1D

    LT_TX_CONFIG_REGISTER = 0x20
    LT_TX_FRAME_REGISTER = 0x24
    LT_RX_STATUS_REGISTER = 0x29
    LT_RX_CONFIG_REGISTER = 0x28
    LT_RX_FRAME_REGISTER = 0x2C
    LT_RX_ERROR_STAT_0 = 0x2A
    LT_RX_ERROR_STAT_1 = 0x2B
    XLA_CONFIG = 0x38
    XLA_TRIG_MASK = 0x39
    XLA_STATUS = 0x3A
    XLA_RD_ADDR = 0x3B
    XLA_RD_PAGE = 0x3C
    XLA_RD_DATA = 0x3D


@dataclass
class AnLtLowLevelInfo:
    base: int
    rx_gtm_base: int
    rx_serdes: int
    tx_gtm_base: int
    tx_serdes: int


async def init(port: GenericL23Port, serdes: int) -> AnLtLowLevelInfo:
    """
    The init function initializes the communication parameters required to read the configuration of a Serializer/Deserializer. 
    It takes in a port object used for communication, and the index of the Serializer/Deserializer to read (serdes). 
    The function returns an object of type AnLtLowLevelInfo, which contains low-level communication information.

    Args:
        port (GenericL23Port): The port object for communication.
        serdes (int): The index of the Serializer/Deserializer to read.


    Returns:
        AnLtLowLevelInfo: An object containing low-level communication information.


    Raises:
        Exception: If there is any error encountered during communication.


    Example:
        >>> port = GenericL23Port()
        >>> inf = await init(port, serdes=0)  # Initialize communication parameters for the first Serializer/Deserializer.
    """
    conn, mid, pid = get_ctx(port)
    inf = await commands.PL1_CFG_TMP(
        conn, mid, pid, serdes, enums.Layer1ConfigType.LL_DEBUG_INFO
    ).get()
    values = inf.values[:5]
    inf = AnLtLowLevelInfo(*values)
    return inf


async def serdes_reset(port: GenericL23Port, serdes: int, inf: t.Optional[AnLtLowLevelInfo] = None) -> None:
    """
    Resets the Serializer/Deserializer specified by serdes.


    The method sets and clears the bit 2 of configuration register GTM_QUAD_GT_CONFIG in order to reset the
    Serializer/Deserializer in question. The port and connection parameters are set through the input `port`. The
    Serializer/Deserializer is specified by the input `serdes`. If the input `inf` is not specified, the `init` method is
    called on the `port` object to initialize the configuration.

    Args:
        port (GenericL23Port): The port object for communication.
        serdes (int): The number of the Serializer/Deserializer to reset.
        inf (AnLtLowLevelInfo, optional): Object with low-level information, which defaults to `None` if not provided.
            This object is used to generate the address of the GTM_QUAD_GT_CONFIG configuration register.

    Returns:
        None.

    Raises:
        Exception: if any async I/O operation encounters an error.

    Examples:
        >>> port = GenericL23Port()
        >>> await serdes_reset(port, serdes=0)  # Reset the first Serializer/Deserializer.
    """
    GTM_QUAD_GT_CONFIG = 0x102
    if inf is None:
        inf = await init(port, serdes)
    conn, mid, pid = get_ctx(port)
    addr = inf.rx_gtm_base + GTM_QUAD_GT_CONFIG + (inf.rx_serdes * 0x40)
    r = commands.PX_RW(conn, mid, pid, 2000, addr)
    v = int((await r.get()).value, 16)
    # Set bit 2
    v |= 1 << 2
    await r.set(value=Hex(f"{v:08X}"))
    # in XOA-Driver V2 `0x` prefix will be drop from the hex strings
    # Clear bit 2
    v &= ~(1 << 2)
    await r.set(value=Hex(f"{v:08X}"))
    return None


async def __get(port: GenericL23Port, serdes: int, reg: AnLtD, inf: t.Optional[AnLtLowLevelInfo] = None) -> int:
    if inf is None:
        inf = await init(port, serdes)
    conn, mid, pid = get_ctx(port)
    addr = inf.base + reg.value
    r = commands.PX_RW(conn, mid, pid, 2000, addr)
    return int((await r.get()).value, 16)


async def __set(port: GenericL23Port, serdes: int, reg: AnLtD, value: int, inf: t.Optional[AnLtLowLevelInfo] = None) -> None:
    if inf is None:
        inf = await init(port, serdes)
    conn, mid, pid = get_ctx(port)
    addr = inf.base + reg.value
    r = commands.PX_RW(conn, mid, pid, 2000, addr)
    await r.set(value=Hex(f"{value:08X}"))
    return None


mode_get = partial(__get, reg=AnLtD.PMD_CONFIG_REGISTER)
mode_set = partial(__set, reg=AnLtD.PMD_CONFIG_REGISTER)

an_status = partial(__get, reg=AnLtD.AN_RX_STATUS_REGISTER)

an_tx_config_get = partial(__get, reg=AnLtD.AN_TX_CONFIG_REGISTER)
an_tx_config_set = partial(__set, reg=AnLtD.AN_TX_CONFIG_REGISTER)

an_rx_config_get = partial(__get, reg=AnLtD.AN_RX_CONFIG_REGISTER)
an_rx_config_set = partial(__set, reg=AnLtD.AN_RX_CONFIG_REGISTER)

an_rx_dme_mv_range_get = partial(__get, reg=AnLtD.AN_RX_DME_MV_RANGE)
an_rx_dme_mv_range_set = partial(__set, reg=AnLtD.AN_RX_DME_MV_RANGE)

an_rx_dme_bit_range_get = partial(__get, reg=AnLtD.AN_RX_DME_BIT_RANGE)
an_rx_dme_bit_range_set = partial(__set, reg=AnLtD.AN_RX_DME_BIT_RANGE)

an_rx_page0_get = partial(__get, reg=AnLtD.AN_RX_PAGE_0_REGISTER)
an_rx_page1_get = partial(__get, reg=AnLtD.AN_RX_PAGE_1_REGISTER)

an_tx_page0_get = partial(__get, reg=AnLtD.AN_TX_PAGE_0_REGISTER)
an_tx_page0_set = partial(__set, reg=AnLtD.AN_TX_PAGE_0_REGISTER)
an_tx_page1_get = partial(__get, reg=AnLtD.AN_TX_PAGE_1_REGISTER)
an_tx_page1_set = partial(__set, reg=AnLtD.AN_TX_PAGE_1_REGISTER)

lt_tx_config_get = partial(__get, reg=AnLtD.LT_TX_CONFIG_REGISTER)
lt_tx_config_set = partial(__set, reg=AnLtD.LT_TX_CONFIG_REGISTER)

lt_rx_config_get = partial(__get, reg=AnLtD.LT_RX_CONFIG_REGISTER)
lt_rx_config_set = partial(__set, reg=AnLtD.LT_RX_CONFIG_REGISTER)

lt_tx_tf_get = partial(__get, reg=AnLtD.LT_TX_FRAME_REGISTER)
lt_tx_tf_set = partial(__set, reg=AnLtD.LT_TX_FRAME_REGISTER)

lt_rx_tf_get = partial(__get, reg=AnLtD.LT_RX_FRAME_REGISTER)

lt_status = partial(__get, reg=AnLtD.LT_RX_STATUS_REGISTER)

lt_rx_error_stat0_get = partial(__get, reg=AnLtD.LT_RX_ERROR_STAT_0)
lt_rx_error_stat1_get = partial(__get, reg=AnLtD.LT_RX_ERROR_STAT_1)

xla_config_get = partial(__get, reg=AnLtD.XLA_CONFIG)
xla_config_set = partial(__set, reg=AnLtD.XLA_CONFIG)

xla_trig_mask_get = partial(__get, reg=AnLtD.XLA_TRIG_MASK)
xla_trig_mask_set = partial(__set, reg=AnLtD.XLA_TRIG_MASK)

xla_status_get = partial(__get, reg=AnLtD.XLA_STATUS)

xla_rd_addr_get = partial(__get, reg=AnLtD.XLA_RD_ADDR)
xla_rd_addr_set = partial(__set, reg=AnLtD.XLA_RD_ADDR)

xla_rd_page_get = partial(__get, reg=AnLtD.XLA_RD_PAGE)
xla_rd_page_set = partial(__set, reg=AnLtD.XLA_RD_PAGE)

xla_rd_data_get = partial(__get, reg=AnLtD.XLA_RD_DATA)


async def lt_prbs(port: GenericL23Port, serdes: int, inf: t.Optional[AnLtLowLevelInfo] = None) -> dict[str, float]:
    """Reads error statistics of an LT PRBS test.

    Args:
        port: A GenericL23Port object.
        serdes: An integer representing the SERDES.
        inf: Optional AnLtLowLevelInfo object.

    Returns:
        A dictionary with total_bits, error_bits, and ber (bit error rate) values.

    Raises:
        Possible exceptions if any async I/O operation encounters an error.
    """ 
    if inf is None:
        inf = await init(port, serdes)

    cfg = await lt_rx_config_get(port, serdes, inf=inf)
    cfg &= ~(3 << 21)  # Clear bit 22-21
    cfg |= 1 << 20  # Set bit 20
    await lt_rx_config_set(port, serdes, inf=inf, value=cfg)  # Trigger PRBS read
    cfg &= ~(1 << 20)  # Clear bit 20
    await lt_rx_config_set(port, serdes, inf=inf, value=cfg)

    # Read the total # bits
    cfg &= ~(3 << 21)  # Clear bit 22-21
    cfg |= 1 << 21
    await lt_rx_config_set(port, serdes, inf=inf, value=cfg)
    v = await lt_rx_error_stat0_get(port, serdes, inf=inf)
    total_bits = v
    v = await lt_rx_error_stat1_get(port, serdes, inf=inf)
    total_bits |= v << 32

    # Read the total # error bits
    cfg &= ~(3 << 21)  # Clear bit 22-21
    cfg |= 2 << 21
    await lt_rx_config_set(port, serdes, inf=inf, value=cfg)
    v = await lt_rx_error_stat0_get(port, serdes, inf=inf)
    error_bits = v
    v = await lt_rx_error_stat1_get(port, serdes, inf=inf)
    error_bits |= v << 32
    error_bits &= 0x0000FFFFFFFFFFFF
    ber = (error_bits) / (total_bits) if total_bits > 0 else float("nan")
    return {"total_bits": total_bits, "error_bits": error_bits, "ber": float(ber)}


async def xla_dump(port: GenericL23Port, serdes: int, inf: t.Optional[AnLtLowLevelInfo] = None) -> t.Dict[str, str]:
    """
    This method takes a GenericL23Port object representing the port for communication, an int serdes representing
    the data serializer for the connection, and an optional AnLtLowLevelInfo object named inf.
    It dumps the 320-bit words in the capture buffer.

    Args:
        port (GenericL23Port): The port for communication.
        serdes (int): The data serializer for the connection.
        inf (AnLtLowLevelInfo, optional): The information object. Defaults to None.

    Returns:
        dict[str,str]: A dictionary of results. The "Trigger Position" key contains the trigger position value
        as a string, and the "Analyzer Status" key contains the status of the analyzer as a string.
    
    Raises:
        Exception: If any async I/O operation encounters an error.

    Examples:
        >>> port= GenericL23Port()
        >>> serdes=1
        >>> inf=None
        >>> result= await xla_dump(port, serdes,inf)
    """
    if inf is None:
        inf = await init(port, serdes)
    result = {}
    trigger_pos, capture_done = await asyncio.gather(
        xla_config_get(port, serdes, inf=inf),
        xla_status_get(port, serdes, inf=inf),
    )
    result["Trigger Position"] = str(trigger_pos)
    result["Analyzer Status"] = str(capture_done)
    if not capture_done:
        result["Data"] = ""
        return result
    data_list = []
    for r in range(256):
        # Set the read address
        await xla_rd_addr_set(port, serdes, inf=inf, value=r)
        for p in range(10):
            # Read the data
            await xla_rd_page_set(port, serdes, inf=inf, value=9 - p)
            d = await xla_rd_data_get(port, serdes, inf=inf)
            data_list.append(f"{d:08X}")
        data_list.append("\n")
    result["Data"] = "".join(data_list)
    return result


async def px_get(port: GenericL23Port, page_address: int, register_address: int) -> t.Tuple[bool, str]:
    """Reads the value of a register located at a specified page address using the PX API.

    Args:
        port (GenericL23Port): A GenericL23Port object representing the port for communication.
        page_address (int): An integer representing the page address of the register to be read.
        register_address (int): An integer representing the address of the register to be read.

    Returns:
        A tuple (bool, str) representing the results of the read operation. The first element of the tuple is a flag 
        indicating whether the read operation was successful or not. The second element represents the value read from 
        the specified register.

    Raises:
        Possible exceptions if any async I/O operation encounters an error.
    """
    resp = await port.transceiver.access_rw(page_address, register_address).get()

    if resp.value.lower().find("dead") != -1:
        return (False, resp.value)
    else:
        return (True, resp.value)


async def px_set(port: GenericL23Port, page_address: int, register_address: int, value: int) -> None:
    """
    Sets a register value in the given page address using the GenericL23Port object for communication.

    Args:
        port (GenericL23Port): The port object for communication.
        page_address (int): The page address of the register.
        register_address (int): The register address.
        value (int): The value to write to the register.

    Returns:
        None.

    Raises:
        Exception: if any async I/O operation encounters an error.

    Examples:
        >>> port= GenericL23Port()
        >>> page_addr=0x1
        >>> reg_addr=0x4
        >>> value= 0x00FF
        >>> await px_set(port,page_addr, reg_addr, value)
    """
    value_hexstr = Hex(f"{value:X}")
    await port.transceiver.access_rw(page_address, register_address).set(value_hexstr)


async def xla_dump_ctrl(port: GenericL23Port, on: bool) -> None:
    """
    Enables or disables XLA mode on the given port by sending a command to set the AN_LT_XLA_MODE configuration
    option. 

    Args:
        port (GenericL23Port): The port object for communication.
        on (bool): Whether to enable XLA mode (True) or disable it (False).

    Returns:
        None.

    Raises:
        Exception: if any async I/O operation encounters an error.

    Examples:
        >>> port = GenericL23Port()
        >>> await xla_dump_ctrl(port, on=False)  # Disable XLA mode on the port.
        >>> await xla_dump_ctrl(port, on=True)   # Enable XLA mode on the port.
    """
    conn, mid, pid = get_ctx(port)
    # await commands.PL1_CFG_TMP(conn, mid, pid, 0, enums.Layer1ConfigType.AN_LT_XLA_MODE).set(values=[int(on)])
    await commands.PL1_CFG_TMP(
        conn, 
        mid, 
        pid, 
        0, 
        enums.Layer1ConfigType.AN_LT_XLA_MODE
    ).set(
        values=[
            enums.OnOff.ON if on else enums.OnOff.OFF
        ]
    )


__all__ = (
    "init",
    "serdes_reset",
    "mode_get",
    "mode_set",
    "lt_prbs",
    "xla_config_get",
    "xla_config_set",
    "xla_dump",
    "xla_rd_addr_get",
    "xla_rd_addr_set",
    "xla_rd_data_get",
    "xla_rd_page_get",
    "xla_rd_page_set",
    "lt_tx_config_get",
    "lt_tx_config_set",
    "xla_status_get",
    "xla_trig_mask_get",
    "xla_trig_mask_set",
    "lt_rx_config_get",
    "lt_rx_config_set",
    "lt_rx_error_stat0_get",
    "lt_tx_tf_get",
    "lt_rx_error_stat1_get",
    "lt_rx_tf_get",
    "lt_status",
    "lt_tx_tf_set",
    "px_get",
    "px_set",
    "xla_dump_ctrl",
    "an_tx_config_get",
    "an_tx_config_set",
    "an_rx_dme_bit_range_get",
    "an_rx_dme_bit_range_set",
    "an_rx_dme_mv_range_get",
    "an_rx_dme_mv_range_set",
    "an_rx_page0_get",
    "an_rx_page1_get",
    "an_status",
    "an_tx_config_get",
    "an_tx_config_set",
    "an_tx_page0_get",
    "an_tx_page0_set",
    "an_tx_page1_get",
    "an_tx_page1_set",
)


"""
Definition of BXMP commands.

All commands in this module are only a stateless declaration of there structures which
define how they will be converted to binary protocol.
"""

# importing commands subsets
from .c_commands import *  # noqa: F403,F401
from .m_commands import *  # noqa: F403,F401
from .m4_commands import *  # noqa: F403,F401
from .m4e_commands import *  # noqa: F403,F401
from .p_commands import *  # noqa: F403,F401
from .p4_commands import *  # noqa: F403,F401
from .p4e_commands import *  # noqa: F403,F401
from .p4g_commands import *  # noqa: F403,F401
from .pc_commands import *  # noqa: F403,F401
from .pd_commands import *  # noqa: F403,F401
from .pe_commands import *  # noqa: F403,F401
from .pec_commands import *  # noqa: F403,F401
from .ped_commands import *  # noqa: F403,F401
from .pef_commands import *  # noqa: F403,F401
from .pf_commands import *  # noqa: F403,F401
from .pl_commands import *  # noqa: F403,F401
from .pm_commands import *  # noqa: F403,F401
from .pp_commands import *  # noqa: F403,F401
from .pr_commands import *  # noqa: F403,F401
from .ps_commands import *  # noqa: F403,F401
from .pt_commands import *  # noqa: F403,F401
from .px_commands import *  # noqa: F403,F401
from .pl1_commands import *  # noqa: F403,F401



from __future__ import annotations
import asyncio
from typing import (
    AsyncIterator,
    Callable,
    Type,
)
from collections import UserDict
from . import registry
from . import exceptions
from ._typings import XoaCommandType
from .protocol.struct_header import ResponseHeader
from .protocol.struct_response import Response


def create_response_obj(cmd: Type[XoaCommandType], header: ResponseHeader, data: bytes) -> Response:
    """Parse bytes retrieved from server to Response structure."""
    return Response(
        class_name=cmd.__name__,
        header=header,
        buffer=data,
        response_struct=getattr(cmd, "GetDataAttr", None)
    )


class CommandsCodeMapper(UserDict):
    data: dict[int, int]

    def add_code(self, req_id: int, cmd_code: int) -> None:
        self.data[req_id] = cmd_code

    def pop_code(self, req_id: int) -> int:
        if command_idx := self.data.pop(req_id, None):
            return command_idx
        raise exceptions.RepeatedRequestID(req_id)


Publisher = Callable[[Response], None]


class PacketsProcessor:
    """Process reading packets from he stream and create a response object for each packet"""

    __slots__ = ("__stream", "__cm_mapper", "__evt_do_job", "__consumer", "__publish",)

    def __init__(self, stream: AsyncIterator[tuple[ResponseHeader, bytes]], publish_func: Publisher) -> None:
        self.__stream = stream
        self.__cm_mapper = CommandsCodeMapper()
        self.__evt_do_job = asyncio.Event()
        self.__consumer: asyncio.Task | None = None
        self.__publish = publish_func

    @property
    def is_running(self) -> bool:
        return not (self.__consumer is None or self.__consumer.done())

    def start(self) -> None:
        if self.is_running:
            return None
        self.__evt_do_job.set()
        self.__consumer = asyncio.create_task(self.__consume())

    def stop(self) -> None:
        if not self.is_running:
            return None
        self.__evt_do_job.clear()
        self.__consumer.cancel()  # type: ignore[ReportOptionalMemberAccess]
        self.__consumer = None

    def register(self, req_id: int, cmd_code: int) -> None:
        self.__cm_mapper.add_code(req_id=req_id, cmd_code=cmd_code)

    async def __task(self, header: ResponseHeader, body_bytes: bytes) -> None:
        command_idx = header.cmd_code if header.is_pushed else self.__cm_mapper.pop_code(header.request_identifier)
        xmc_type = registry.get_command(command_idx)
        response = create_response_obj(xmc_type, header, body_bytes)
        self.__publish(response)

    async def __consume(self) -> None:
        async for header, body_bytes in self.__stream:
            asyncio.create_task(self.__task(header, body_bytes))
            if not self.__evt_do_job.is_set():
                return None



from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    P4E_ASSIGN,
    P4E_AVAILABLE,
    P4E_ALLOCATE,
    P4E_ALLOCATION_INFO,
)


class PacketEngine:
    """L47 port's packet engine."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.assign = P4E_ASSIGN(conn, module_id, port_id)
        """Representation of P4E_ASSIGN"""
        self.available = P4E_AVAILABLE(conn, module_id, port_id)
        """Representation of P4E_AVAILABLE"""
        self.allocate = P4E_ALLOCATE(conn, module_id, port_id)
        """Representation of P4E_ALLOCATE"""
        self.allocation_info = P4E_ALLOCATION_INFO(conn, module_id, port_id)
        """Representation of P4E_ALLOCATION_INFO"""
